#include "StdAfx.h"
#include "LeoHelpers.h"

bool NLeoHelpers::LeetRegQueryValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, REGSAM sam,
									   DWORD *pdwType, void **plpData, DWORD *pcbData)
{
	// Documentation doesn't state that registry functions call SetLastError().
	// We need to preserve the error values they return through clean-up anyway.
	LONG regRes = ERROR_SUCCESS;

	HKEY hKeyToReadValueFrom;

	DWORD dwPcdDataStandin; // Used for getting the registry size if the user doesn't care about it.

	if (NULL == pcbData)
	{
		pcbData = &dwPcdDataStandin;
	}

	// szKeyPath may be NULL in which case another handle like hKeyParent is openned (and must be closed).
	regRes = RegOpenKeyEx(hKeyParent, szKeyPath, 0, sam | KEY_QUERY_VALUE, &hKeyToReadValueFrom);

	if (ERROR_SUCCESS == regRes)
	{
		if (NULL == plpData)
		{
			// They just want the type and/or the data size.
			regRes = RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, NULL, pcbData);
		}
		else
		{
			// We add 8 bytes of zero padding to the end of any buffer we return. This is because
			// the registry does not check inputs to ensure they are the correct length for the
			// datatype and also does not ensure that strings are null-terminated and multi-strings are
			// double-null-terminated. Rather than add checks into all functions which call this one,
			// we pad with as many zero bytes as are needed to correct anything we might read.
			// 8 means a QUADDWORD that was zero length in the registry would still be safe.
			// 8 is also more than enough for double-null-terminator of a Unicode multi-string.
			const DWORD dwZeroPadding = 8;

			BYTE *allocatedBuffer = NULL;
			DWORD dwAllocatedBufferSize = 0; // For HKEY_PERFORMANCE_DATA, really.

			// Get the required buffer size by giving NULL for data buffer.
			regRes = RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, NULL, pcbData);

			if (ERROR_SUCCESS == regRes)
			{
				dwAllocatedBufferSize = *pcbData;
				allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
			}
			else if (ERROR_MORE_DATA == regRes)
			{
				// Most likely tried to read HKEY_PERFORMANCE_DATA. *pcdData is undefined at this point.
				*pcbData = 1024; // Must be a number x such that (Int(x/2) > 0)
				dwAllocatedBufferSize = *pcbData;
				allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
			}
			else
			{
				*pcbData = 0;
				// leave allocatedBuffer NULL so the while loop does not execute.
			}

			while (NULL != allocatedBuffer)
			{
				regRes = RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, allocatedBuffer, pcbData);

				if (ERROR_SUCCESS == regRes)
				{
					// Zero out the padding we added.
					ZeroMemory(allocatedBuffer + dwAllocatedBufferSize, dwZeroPadding);

					*plpData = allocatedBuffer;
					allocatedBuffer = NULL;
					break;
				}
				else if (ERROR_MORE_DATA == regRes)
				{
					// Try a larger buffer.
					delete [] allocatedBuffer;
					*pcbData = dwAllocatedBufferSize + (dwAllocatedBufferSize/2);
					dwAllocatedBufferSize = *pcbData;
					allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
				}
				else
				{
					*pcbData = 0;
					delete [] allocatedBuffer;
					allocatedBuffer = NULL;
					break;
				}
			}
		}

		LONG closeRes = RegCloseKey(hKeyToReadValueFrom);
		if (ERROR_SUCCESS == regRes) { regRes = closeRes; }
	}

	if (ERROR_SUCCESS != regRes)
	{
		SetLastError(regRes);
	}

	return(ERROR_SUCCESS == regRes);
}

// Returns false on failure. Use GetLastError().
// Uses LeetRegQueryValue() to get you a TCHAR string without having to worry about buffers, types and so on.
bool NLeoHelpers::LeetRegQueryStringValue(HKEY hKeyParent, const TCHAR *szKeyPath, const TCHAR *szValueName, REGSAM sam, std::basic_string<TCHAR> *pString)
{
	DWORD dwErrToSet = ERROR_SUCCESS;

	DWORD dwType;
	void *pData;

	pString->erase();

	if (! LeetRegQueryValue(hKeyParent, szKeyPath, szValueName, sam, &dwType, &pData, NULL) )
	{
		dwErrToSet = GetLastError();
	}
	else
	{
		if (REG_SZ != dwType)
		{
			dwErrToSet = ERROR_INVALID_DATA;
		}
		else
		{
			(*pString) = reinterpret_cast<TCHAR*>(pData);
		}

		delete [] pData;
	}

	if (ERROR_SUCCESS != dwErrToSet)
	{
		SetLastError(dwErrToSet);
	}

	return(ERROR_SUCCESS == dwErrToSet);
}

TCHAR *NLeoHelpers::GetParentPath(const TCHAR *path)
{
	TCHAR *result = NULL;

	if (NULL != path)
	{
		const TCHAR *lastBackslash  = _tcsrchr(path, _T('\\'));
		const TCHAR *lastFrontslash = _tcsrchr(path, _T('/') );
		const TCHAR *lastSlash = NULL;

		if ((NULL != lastBackslash) && (NULL != lastFrontslash))
		{
			if (lastBackslash > lastFrontslash)
			{
				lastSlash = lastBackslash;
			}
			else
			{
				lastSlash = lastFrontslash;
			}
		}
		else if (NULL != lastBackslash)
		{
			lastSlash = lastBackslash;
		}
		else if (NULL != lastFrontslash)
		{
			lastSlash = lastFrontslash;
		}

		if (NULL == lastSlash)
		{
			result = new TCHAR[1];
			result[0] = _T('\0');
		}
		else
		{
			size_t resSize = (lastSlash - path) + 1;
			result = new TCHAR[resSize];
			NLeoHelpers::StringCopy(result, path, resSize);
		}
	}

	return(result);
}

int NLeoHelpers::StringCopy(TCHAR *dest, const TCHAR *source, int destSizeChars)
{
	int idx = 0;

	if (destSizeChars > 0)
	{
		while ((_T('\0') != source[idx]) && (destSizeChars > 1))
		{
			dest[idx] = source[idx];
			idx++;
			destSizeChars--;
		}

		dest[idx] = _T('\0');
	}

	return(idx);
}

// _snprintf doesn't always null term the destination... sigh
// Returns the index of the null terminator.
int NLeoHelpers::StringFormat(TCHAR *buffer, size_t bufferSizeChars, const TCHAR *format, ...)
{
	va_list args;
	va_start(args, format);
	int result = VStringFormat(buffer, bufferSizeChars, format, args);
	va_end(args);

	return(result);
}

// _vsnprintf doesn't always null term the destination... sigh
// Returns the index of the null terminator.
int NLeoHelpers::VStringFormat(TCHAR *buffer, size_t bufferSizeChars, const TCHAR *format, va_list args)
{
	int result = 0;

	if (bufferSizeChars > 0)
	{
		result = _vsntprintf(buffer, bufferSizeChars, format, args);

		if (0 > result)
		{
			result = bufferSizeChars - 1; // LD 20/Mar/2000
			buffer[result] = _T('\0');
		}
	}

	return(result);
}

// delete[] the result.
// Returns NULL on failure.
TCHAR *NLeoHelpers::StringAllocAndFormat(const TCHAR *format, ...)
{
	va_list args;
	va_start(args, format);
	TCHAR *result = VStringAllocAndFormat(format, args);
	va_end(args);

	return(result);
}

// delete[] the result.
// Returns NULL on failure.
TCHAR *NLeoHelpers::VStringAllocAndFormat(const TCHAR *format, va_list args)
{
	size_t bufferSize = 1024;

	TCHAR *szResult = new TCHAR[bufferSize];

	while(true)
	{
		int formatRes = _vsntprintf(szResult, bufferSize, format, args);

		// formatRes negative on error; (bufferSize == formatRes) should be redundant for _vsntprintf.

		if ((0 > formatRes) || (bufferSize == formatRes))
		{
			delete [] szResult;
			bufferSize *= 2;
			szResult = new TCHAR[bufferSize];
		}
		else
		{
			break;
		}
	}

	return(szResult);
}

TCHAR *NLeoHelpers::JoinPaths(const TCHAR *first, const TCHAR *second)
{
	TCHAR *result = NULL;

	int firstLen = _tcslen(first);
	int length = firstLen + _tcslen(second) + 1;

	// LD 10/Jan/2002 Now checks that the start of second isn't a slash either.

	bool addSlash = ((firstLen > 0)			&& (_T('\\') != first[firstLen - 1]) && (_T('/') != first[firstLen - 1])\
				&&	 (_tcslen(second) > 0)	&& (_T('\\') != second[0])			 && (_T('/') != second[0]));

	if (addSlash)
	{
		length++; // Space for the extra backslash.
	}

	result = new TCHAR[length];

	_tcscpy(result, first);

	if (addSlash)
	{
		result[firstLen] = _T('\\');
		_tcscpy(result + firstLen + 1, second);
	}
	else
	{
		_tcscpy(result + firstLen, second);
	}

	return(result);
}

WCHAR *NLeoHelpers::LeetMBtoWC(const char *mbString)
{
	WCHAR *result = NULL;

	int wcBufSize = MultiByteToWideChar(CP_ACP, 0, mbString, -1, NULL, 0);

	result = new WCHAR[wcBufSize];

	// Hopefully this includes the '\0' in the copy and the count of characters,
	// else this will not work with empty strings.
	if (0 == MultiByteToWideChar(CP_ACP, 0, mbString, -1, result, wcBufSize))
	{
		delete [] result;
		result = NULL;
	}

	return(result);
}

char *NLeoHelpers::LeetWCtoMB(const WCHAR *wcString)
{
	char *result = NULL;

	int mbBufSize = WideCharToMultiByte(CP_ACP, 0, wcString, -1, NULL, 0, NULL, NULL);

	result = new char[mbBufSize];

	// Hopefully this includes the '\0' in the copy and the count of characters,
	// else this will not work with empty strings.
	if (0 == WideCharToMultiByte(CP_ACP, 0, wcString, -1, result, mbBufSize, NULL, NULL))
	{
		delete [] result;
		result = NULL;
	}

	return(result);
}

const TCHAR *NLeoHelpers::GetLastPathPart(const TCHAR *szIn)
{
	if (NULL == szIn)
	{
		return(szIn);
	}

	size_t len = _tcslen(szIn);

	if (1 >= len)
	{
		return(szIn);
	}

	const TCHAR *szOut = szIn + (len - 1);

	// Ignore any slashes at the very end of the string.
	while (szIn < szOut && (*szOut == _T('\\') || *szOut == _T('/')))
	{
		szOut--;
	}

	// Stop at the next slash we find.
	while (szIn < szOut && (*szOut != _T('\\') && *szOut != _T('/')))
	{
		szOut--;
	}

	if (*szOut == _T('\\') || *szOut == _T('/'))
	{
		szOut++;
	}

	return(szOut);
}
